package mr

import (
	"fmt"
	"log"
	"net"
	"net/http"
	"net/rpc"
	"os"
)

type Coordinator struct {
	// Your definitions here.
	mapTask    []MapTask
	recudeTask []ReduceTask

	mapTaskQueue    chan int
	reduceTaskQueue chan int
	running         chan bool

	lock chan bool
}

// Your code here -- RPC handlers for the worker to call.
func (c *Coordinator) GetJob(args *Empty, reply *Task) error {
	select {
	case id := <-c.mapTaskQueue:
		reply.MapTask = &c.mapTask[id]
		reply.MapTask.TaskId = id
		reply.MapTask.State = PROGRESS

		reply.TaskType = TASK_MAP
	case id := <-c.reduceTaskQueue:
		reply.ReduceTask = &c.recudeTask[id]
		reply.ReduceTask.TaskId = id
		reply.ReduceTask.State = PROGRESS

		entries, err := os.ReadDir(".")
		if err != nil {
			log.Fatal("read dir error")
		}
		var mapId, reduceId int
		for _, entry := range entries {

			n, _ := fmt.Sscanf(entry.Name(), "mr-%d-%d", &mapId, &reduceId)
			if n == 2 && reduceId == reply.ReduceTask.TaskId {
				// log.Printf("reduce task %d find file %s", reply.ReduceTask.TaskId, entry.Name())
				reply.ReduceTask.Fnames = append(reply.ReduceTask.Fnames, entry.Name())
			}
		}

		reply.TaskType = TASK_REDUCE
	}
	return nil
}

func (c *Coordinator) SubmitJob(args *Task, reply *Empty) error {
	<-c.lock
	defer func() { c.lock <- true }()

	switch args.TaskType {
	case TASK_MAP:
		// log.Print("submit map task")
		c.mapTask[args.MapTask.TaskId].State = COMPELETED
		for _, task := range c.mapTask {
			if task.State != COMPELETED {
				return nil
			}
		}

		log.Print("start reduce task")
		for _, task := range c.recudeTask {
			c.reduceTaskQueue <- task.TaskId
		}
	case TASK_REDUCE:
		// log.Print("submit reduce task")
		c.recudeTask[args.ReduceTask.TaskId].State = COMPELETED
		for _, task := range c.recudeTask {
			if task.State != COMPELETED {
				return nil
			}
		}

		log.Print("end")
		// close(c.reduceTaskQueue)
		// close(c.mapTaskQueue)
		<-c.running
		c.running <- false
	}
	return nil
}

// an example RPC handler.
//
// the RPC argument and reply types are defined in rpc.go.
func (c *Coordinator) Example(args *ExampleArgs, reply *ExampleReply) error {
	reply.Y = args.X + 1
	return nil
}

// start a thread that listens for RPCs from worker.go
func (c *Coordinator) server() {
	rpc.Register(c)
	rpc.HandleHTTP()
	//l, e := net.Listen("tcp", ":1234")
	sockname := coordinatorSock()
	os.Remove(sockname)
	l, e := net.Listen("unix", sockname)
	if e != nil {
		log.Fatal("listen error:", e)
	}
	go http.Serve(l, nil)
}

// main/mrcoordinator.go calls Done() periodically to find out
// if the entire job has finished.
func (c *Coordinator) Done() bool {
	ret := false

	// Your code here.
	state := <-c.running
	ret = !state
	c.running <- state

	return ret
}

// create a Coordinator.
// main/mrcoordinator.go calls this function.
// nReduce is the number of reduce tasks to use.
func MakeCoordinator(files []string, nReduce int) *Coordinator {
	c := Coordinator{
		mapTask:         make([]MapTask, len(files)),
		recudeTask:      make([]ReduceTask, nReduce),
		mapTaskQueue:    make(chan int, len(files)),
		reduceTaskQueue: make(chan int, nReduce),
		running:         make(chan bool, 1),
		lock:            make(chan bool, 1),
	}

	// Your code here.
	c.running <- true
	c.lock <- true

	for i, file := range files {
		c.mapTask[i] = MapTask{
			Fname:     file,
			State:     IDLE,
			NReduce:   nReduce,
			TaskId:    i,
			MachineId: -1,
		}
	}

	for i := range nReduce {
		c.recudeTask[i] = ReduceTask{
			Fnames:    make([]string, 0),
			State:     IDLE,
			TaskId:    i,
			MachineId: -1,
		}
	}

	for _, task := range c.mapTask {
		c.mapTaskQueue <- task.TaskId
	}

	c.server()
	return &c
}
